/* $Id: mman.h,v 1.7 1996/04/25 06:13:15 davem Exp $ */
#ifndef __SPARC_MMAN_H__
#define __SPARC_MMAN_H__

/* SunOS'ified... */

#define PROT_READ	0x1		/* page can be read */
#define PROT_WRITE	0x2		/* page can be written */
#define PROT_EXEC	0x4		/* page can be executed */
#define PROT_NONE	0x0		/* page can not be accessed */

#define MAP_SHARED	0x01		/* Share changes */
#define MAP_PRIVATE	0x02		/* Changes are private */
#define MAP_TYPE	0x0f		/* Mask for type of mapping */
#define MAP_FIXED	0x10		/* Interpret addr exactly */
#define MAP_ANONYMOUS	0x20		/* don't use a file */
#define MAP_RENAME      MAP_ANONYMOUS   /* In SunOS terminology */
#define MAP_NORESERVE   0x40            /* don't reserve swap pages */
#define MAP_INHERIT     0x80            /* SunOS doesn't do this, but... */
#define MAP_LOCKED      0x100           /* lock the mapping */
#define _MAP_NEW        0x80000000      /* Binary compatibility is fun... */

#define MAP_GROWSDOWN	0x0100		/* stack-like segment */
#define MAP_DENYWRITE	0x0800		/* ETXTBSY */
#define MAP_EXECUTABLE	0x1000		/* mark it as a executable */

#define MS_ASYNC	1		/* sync memory asynchronously */
#define MS_INVALIDATE	2		/* invalidate the caches */
#define MS_SYNC		4		/* synchronous memory sync */

#define MCL_CURRENT     0x2000          /* lock all currently mapped pages */
#define MCL_FUTURE      0x4000          /* lock all additions to address space */

/* XXX Need to add flags to SunOS's mctl, mlockall, and madvise system
 * XXX calls.
 */

/* SunOS sys_mctl() stuff... */
#define MC_SYNC         1  /* Sync pages in memory with storage (usu. a file) */
#define MC_LOCK         2  /* Lock pages into core ram, do not allow swapping of them */
#define MC_UNLOCK       3  /* Unlock pages locked via previous mctl() with MC_LOCK arg */
#define MC_LOCKAS       5  /* Lock an entire address space of the calling process */
#define MC_UNLOCKAS     6  /* Unlock entire address space of calling process */

/* compatibility flags */
#define MAP_ANON	MAP_ANONYMOUS
#define MAP_FILE	0

#endif /* __SPARC_MMAN_H__ */
